-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.sha3_blake_package.all;

entity blake_control is		
	generic ( 	 	 
		VERSION : integer := SHA3_ROUND3;
		HS : integer := 256
	);
	port (					
		rst			: in std_logic;
		clk			: in std_logic;
		src_ready	: in std_logic;
		src_read	: out std_logic;
		dst_ready 	: in std_logic;
		dst_write	: out std_logic;
		
		load_c, ec	: out std_logic;		
		dt, eth, etl: out std_logic;
		
		stbot               : out std_logic;
		zc0, zcRemainder    : in std_logic;		
        extra_block	        : in std_logic;		
		final_segment 		: in std_logic;
		
		msg_wr, state_wr    : out std_logic;
        h_wr, sf, smux      : out std_logic;
		round 		        : out std_logic_vector(7 downto 0);
        
        -- pad
		last_word       : out std_logic;
		en_lb           : out std_logic;
        clr_lb          : out std_logic;
		sel_pad         : out std_logic_vector(1 downto 0);
		spos            : out std_logic_vector(1 downto 0);	 
		comp_rem_e0     : in std_logic;
        comp_lb_e0      : in std_logic;    
        comp_rem_mt_block  : in std_logic
	);				 
end blake_control;

architecture struct of blake_control is				   	 
	constant roundnr 			: integer := get_roundnr( HS, VERSION )*16;
	constant log2roundnrzeros	: std_logic_vector(log2( roundnr )-1 downto 0) := (others => '0');
	
	type fsm_type is (reset, wait_for_header, read_header, load, initialization, hash, finalization, output_data_stall, output_data );
	signal nstate, cs : fsm_type;
	
	signal pc : std_logic_vector(log2(roundnr)-1 downto 0) := (others =>'0');
	signal lpc, epc : std_logic;
	signal set_fs, clr_fs, fs_true : std_logic;	-- final segment sr
	signal init_done, set_init, clr_init : std_logic; -- init sr
	
	signal zpc0, zpc7, zpc3, zpc15, zpcRound : std_logic;
	signal zpc13, zpc14 : std_logic;
	signal set_stbot, clr_stbot, stbot_s : std_logic;
	signal extra, set_extra, clr_extra : std_logic;
	
	-- output registers;
	signal h_wr_sig, state_wr_sig, load_c_sig, ec_sig, dt_sig, eth_sig, etl_sig : std_logic;
	signal msg_wr_sig, sf_sig, smux_sig, dst_write_sig, src_read_sig : std_logic;
	
	-- PAD
	signal set_start_pad, clr_start_pad, start_pad : std_logic;			 
	signal set_full_block, clr_full_block, full_block : std_logic;
	signal set_empty_block, clr_empty_block, empty_block : std_logic;
	signal spos_sig : std_logic_vector(1 downto 0);
begin
	sr_final_segment : sr_reg 
		port map ( rst => rst, clk => clk, set => set_fs, clr => clr_fs, output => fs_true);
		
	sr_stbot : sr_reg 
		port map ( rst => rst, clk => clk, set => set_stbot, clr => clr_stbot, output => stbot_s);
	
	
	sr_init : sr_reg 
		port map ( rst => rst, clk => clk, set => set_init, clr => clr_init, output => init_done);
	
	sr_extra : sr_reg 
		port map ( rst => rst, clk => clk, set => set_extra, clr => clr_extra, output => extra);
	
	round_counter_gen : countern generic map ( N => 8 ) 
		port map ( clk => clk, rst => rst, load => lpc, en => epc, input => log2roundnrzeros, output => pc);
	
	
	-- state process
	cstate_proc : process ( clk )
	begin
		if rising_edge( clk ) then 
			if rst = '1' then
				cs <= reset;
			else
				cs <= nstate;
			end if;	   
		end if;
	end process;
	
	zpc0 <= '1' when pc(3 downto 0) = 0 else '0';
	zpc3 <= '1' when pc(3 downto 0) = 3 else '0';
	zpc7 <= '1' when pc(3 downto 0) = 7 else '0';
	zpc13 <= '1' when pc(3 downto 0) = 13 else '0';
	zpc14 <= '1' when pc(3 downto 0) = 14 else '0';
	zpc15 <= '1' when pc(3 downto 0) = 15 else '0';
	zpcRound <= '1' when pc = roundnr-1 else '0';
	
	nstate_proc : process ( cs, src_ready, dst_ready, zc0, zcRemainder, final_segment, fs_true, zpc3, zpc7, zpc15, zpcRound, init_done, extra, comp_rem_e0 )
	begin					  
		if ( cs = reset ) then
			nstate <= wait_for_header;
		elsif ( cs = wait_for_header ) then
			if ( src_ready = '0' and final_segment = '0' ) then
				nstate <= load;							
			elsif ( src_ready = '0' and final_segment = '1' ) then
				nstate <= read_header;
			else
				nstate <= wait_for_header;
			end if;			  
		elsif ( cs = read_header ) then
			nstate <= load;
		elsif ( cs = load ) then			
			if (src_ready = '0' and zpc15 = '1') or (src_ready = '1'  and zpc15 = '1' and fs_true = '1' and comp_rem_e0 = '1' and zc0 = '1') then
				if init_done = '1' then
					nstate <= finalization;
				else
					nstate <= initialization;
				end if;			
			else
				nstate <= load;
			end if;
		elsif ( cs = initialization ) then
			if ( zpc3 = '1' ) then
				nstate <= hash;
			else
				nstate <= initialization;
			end if;
		elsif ( cs = finalization ) then
			if ( zpc3 = '1' ) then
				nstate <= hash;				
			else
				nstate <= finalization;
			end if;
		elsif ( cs = hash ) then
			if ( zpcRound = '1' ) then 
				if ( fs_true = '1' and zc0 = '1' and zcRemainder = '0' and extra = '0') then
					nstate <= output_data;
				elsif ( fs_true = '0' and zc0 = '1' ) then
					nstate <= wait_for_header;
				else
					nstate <= load;
				end if;
			else
				nstate <= hash;
			end if;		  
		elsif ( cs = output_data_stall) then
			nstate <= output_data;			
		elsif ( cs = output_data ) then
			if ( zpc7 = '1' and dst_ready = '0') then
				nstate <= wait_for_header;	  
			elsif ( dst_ready = '0' ) then
				nstate <= output_data_stall;
			else
				nstate <= output_data;
			end if;
		end if;
	end process;
	
	-- FIFO	  
	sr_empty_block : sr_reg 
		port map ( rst => rst, clk => clk, set => set_empty_block, clr => clr_empty_block, output => empty_block);	
	set_empty_block <= '1' when (cs = hash and zpcRound = '1' and zc0 = '1' and fs_true = '1' and comp_rem_e0 = '1') else '0';
	clr_empty_block <= '1' when (cs = finalization  or cs = initialization or cs = reset) else '0';
	
	sr_full_block : sr_reg 
		port map ( rst => rst, clk => clk, set => set_full_block, clr => clr_full_block, output => full_block);	
	clr_full_block <= '1' when (cs = reset) or (cs = wait_for_header) or (cs = hash and zpcRound = '1' and zc0 = '1' and fs_true = '1') else '0';	  -- assume the first block is not full always
	set_full_block <= '1' when (cs = finalization) else '0';
	
	src_read_sig <= '1' when 
		(cs = wait_for_header and src_ready = '0') or
		(cs = load and ((src_ready = '0' and zc0 = '0') or (src_ready = '0' and fs_true = '1' and zc0 = '1' and comp_rem_e0 = '0'))) else '0';
	
	dst_write_sig <= '1' when (cs = output_data and dst_ready = '0') else '0';		
	load_c_sig <= '1' when (cs = wait_for_header) else '0';
	etl_sig <= '1' when (cs = wait_for_header) else '0';

	set_fs <= '1' when (src_ready = '0' and cs = wait_for_header and final_segment = '1') else '0';
	clr_fs <= '1' when (cs = output_data or cs = reset) else '0';					  
		
	ec_sig <= '1' when (cs = load and zc0 = '0' and src_ready = '0' and zpc15 = '1')  else '0';
	eth_sig <= '1' when (cs = load and zc0 = '0'  and src_ready = '0' and zpc15 = '1') else '0';  	 
	dt_sig <= '1' when (
		(cs = output_data_stall) or
		(cs = reset) or 
		(cs = load and zpc14 = '1' and extra = '0' and zc0 = '1' and fs_true = '1' and init_done = '1' and empty_block = '1') or
		(cs = hash and zpc15 = '1' and fs_true = '1' and extra = '0' and comp_rem_e0 = '1')
	) else '0';
	
	
	msg_wr_sig <= '1' when (cs = load and src_ready = '0') or ( cs = load and src_ready = '1' and fs_true = '1' and comp_rem_e0 = '1') else '0';
	state_wr_sig <= '1' when (cs = initialization or cs = hash or cs = finalization ) else '0';
	h_wr_sig <= '1' when (cs = initialization or cs = finalization ) else '0';
	smux_sig <= '1' when (cs = initialization or cs = finalization or cs = output_data ) else '0';
	sf_sig <= '1' when (cs = initialization) else '0';
	
	set_extra <= '1' when (cs = read_header) and
		(
			(fs_true = '1' and comp_rem_mt_block = '1' and extra = '0') or
			(fs_true = '1' and comp_rem_e0 = '1' and extra = '0')
		) 
		else '0';
	clr_extra <= '1' when (cs = hash and zpcRound = '1' and extra = '1' and zc0 = '1' and comp_rem_e0 = '1') else '0';
	
	set_stbot <= '1' when ((cs = load) and zc0 = '1' and zcRemainder = '1' and zpc7 = '1') else '0';	
	clr_stbot <= '1' when (cs = output_data ) else '0';
	-- pc 
	
	epc <= '1' when (
		( cs = load and src_ready = '0') or
		( cs = load and src_ready = '1' and fs_true = '1' and comp_rem_e0 = '1' and zc0 = '1') or 
		( cs = initialization ) or
		( cs = hash ) or
		( cs = finalization ) or
		( cs = output_data and dst_ready = '0'))
		else '0';
	lpc <= '1' when (
		( cs = load and src_ready = '0' and zpc15 = '1') or
		( cs = load and src_ready = '1' and zpc15 = '1' and fs_true = '1' and comp_rem_e0 = '1' and zc0 = '1') or
		( cs = initialization  and zpc3 = '1') or
		( cs = hash  and zpcRound = '1') or
		( cs = finalization  and zpc3 = '1') or
		( cs = output_data and dst_ready = '0' and zpc7 = '1'))
		else '0';
				
	-- init sr
	set_init <= '1' when (cs = initialization) else '0';
	clr_init <= '1' when (cs = output_data or cs = reset) else '0';
		
		
	output_reg : process( clk ) 
	begin
		if rising_edge( clk ) then
			stbot <= stbot_s;
			dst_write <= dst_write_sig;						
			eth <= eth_sig;			
			state_wr <= state_wr_sig;
			h_wr <= h_wr_sig;
			
			sf <= sf_sig;			
			dt <= dt_sig;	 			
			ec <= ec_sig;						
		end if;
	end process;	  
	
	smux <= smux_sig;
	msg_wr <= msg_wr_sig;	
	load_c <= load_c_sig;
	etl <= etl_sig; 
	round <= pc;
	src_read <= src_read_sig;	   
		   		   
	-- ////////////////				   
	-- PAD controls	
	sr_pos1 :  sr_reg 
        port map ( rst => rst, clk => clk, set => set_start_pad, clr => clr_start_pad, output => start_pad);
	spos_sig <= "00" when (cs = load and fs_true = '1' and comp_lb_e0 = '1' and start_pad = '0' and full_block = '0' and zc0 = '1') else	-- select '0'
			"01" when (cs = load and fs_true = '1' and comp_lb_e0 = '1' and start_pad = '1' and full_block = '0' and zc0 = '1') else 	-- select start pad
			"11"; 
	spos <= spos_sig;
	set_start_pad <= '1' when (cs = wait_for_header and final_segment = '1' ) else '0';
	clr_start_pad <= '1' when (cs = load and fs_true = '1' and comp_lb_e0 = '1' and start_pad = '1' and epc = '1' and zc0 = '1') else '0';
							   	
	sel_pad(0) <= '1' when (cs = load and zpc15 = '1' and extra = '0' and fs_true = '1' and full_block = '0' and zc0 = '1') else '0';		
	sel_pad(1) <= '1' when (cs = load and (zpc14 = '1' or zpc15 = '1') and extra = '0' and fs_true = '1' and full_block = '0' and zc0 = '1') else '0';
	-- last word indicator
	last_word <= '1' when (cs = load and zpc13 = '1' and fs_true = '1' and extra = '0') else '0';
	
	-- last block counter
	en_lb <= '1' when (cs = load and fs_true = '1' and comp_lb_e0 = '0' and src_ready = '0' and zc0 = '1') else '0';
	clr_lb <= '1' when (cs = load and fs_true = '1' and comp_lb_e0 = '1' and src_ready = '0' and epc = '1' and zc0 = '1') else '0';			

end struct;